@@ -2,18 +2,20 @@
 from __future__ import print_function  # Must be at the beginning of the file
 
 import argparse
+import io
 import os
 import os.path
 import re
-import shlex
 import sys
-import xym
 from collections import Counter
-from subprocess import Popen, PIPE
 
 import requests
+from pyang import plugin, error
+from pyang.plugins.name import emit_name
 from requests.packages.urllib3 import disable_warnings
 
+from .yangParser import create_context
+
 __author__    = 'jmedved@cisco.com, calle@tail-f.com, bclaise@cisco.com, einarnn@gmail.com'
 __copyright__ = "Copyright(c) 2015, 2016, 2017, 2020 Cisco Systems, Inc."
 __license__   = "New-style BSD"
@@ -69,11 +71,11 @@ class YangModuleExtractor:
     MODULE_STATEMENT = re.compile('''^[ \t]*(sub)?module +(["'])?([-A-Za-z0-9]*(@[0-9-]*)?)(["'])? *\{.*$''')
     PAGE_TAG = re.compile('.*\[Page [0-9]*\].*')
     CODE_ENDS_TAG = re.compile('^[} \t]*<CODE ENDS>.*$')
-    CODE_BEGINS_TAG = re.compile('^[ \t]*<CODE BEGINS>( *file(\W+"(.*)")?)?[ \t]*$')
+    CODE_BEGINS_TAG = re.compile('^[ \t\r\n]*<CODE BEGINS>( *file(?:(\W+"(.*)")|(?:\W+"(.*)))?)?[ \t\r\n]*$')
     EXAMPLE_TAG = re.compile('^(example-)')
 
     def __init__(self, src_id, dst_dir, strict=True, strict_examples=True, strict_name=False, add_line_refs=False,
-                 debug_level=0):
+                 debug_level=0, skip_modules=None, parse_only_modules=None):
         """
         Initializes class-global variables.
         :param src_id: text string containing the draft or RFC text from which YANG
@@ -84,6 +86,8 @@ class YangModuleExtractor:
         :param strict_examples: Only output valid examples when in strict mode
         :param strict_name: enforce name from module
         :param debug_level: If > 0 print some debug statements to the console
+        :param skip_modules: it will skip modules in this list
+        :param parse_only_modules: it will parse only modules in this list
         :return:
         """
         self.src_id = src_id
@@ -93,6 +97,8 @@ class YangModuleExtractor:
         self.strict_name = strict_name
         self.add_line_refs = add_line_refs
         self.debug_level = debug_level
+        self.skip_modules = [] if skip_modules is None else skip_modules 
+        self.parse_only_modules = [] if parse_only_modules is None else parse_only_modules
         self.max_line_len = 0
         self.extracted_models = []
 
@@ -140,10 +146,12 @@ class YangModuleExtractor:
                         bt = m.group(1)
                         continue
 
+        if mrev is not None and mrev.rstrip() != '':
+            mrev = '@' + mrev
         if bt != '':
-            return mname + '@' + mrev + ' (belongs-to {})'.format(bt)
+            return mname + mrev + ' (belongs-to {})'.format(bt)
 
-        return mname + '@' + mrev
+        return mname + mrev
 
     def get_extracted_models(self, force_revision_pyang, force_revision_regexp):
         if force_revision_pyang or force_revision_regexp:
@@ -151,21 +159,57 @@ class YangModuleExtractor:
             models.extend(self.extracted_models)
             for model in models:
                 if force_revision_pyang:
-                    command = '/usr/local/bin/pyang -f name-revision "' + self.dst_dir + '/' + model + '"'
-                    proc = Popen(shlex.split(command), stdout=PIPE, stderr=PIPE)
-                    out, err = proc.communicate()
+                    if isinstance(self.dst_dir, list):
+                        path = ':'.join(self.dst_dir)
+                    else:
+                        path = self.dst_dir
+                    ctx = create_context(path)
+                    ctx.opts.print_revision = True
+                    for p in plugin.plugins:
+                        p.setup_ctx(ctx)
+                    with open(self.dst_dir + '/' + model, 'r', encoding="utf-8") as yang_file:
+                        module = yang_file.read()
+                        m = ctx.add_module(self.dst_dir + '/' + model, module)
+                        if m is None:
+                            m = []
+                        else:
+                            m = [m]
+                    ctx.validate()
+
+                    f = io.StringIO()
+                    emit_name(ctx, m, f)
+                    out = f.getvalue()
+
                     if out.rstrip() == '':
+
+                        def __print_pyang_output(ctx):
+                            err = ''
+                            for (epos, etag, eargs) in ctx.errors:
+                                elevel = error.err_level(etag)
+                                if error.is_warning(elevel):
+                                    kind = "warning"
+                                else:
+                                    kind = "error"
+
+                                err += str(epos) + ': %s: ' % kind + \
+                                       error.err_to_str(etag, eargs) + '\n'
+                            return err
+
+                        err = __print_pyang_output(ctx)
                         if err:
-                            self.error('extracting revision from file with: pyang -f name-revision ' + self.dst_dir +
+                            self.error('extracting revision from file with pyang ' + self.dst_dir +
                                        '/' + model + ' has following errors:\n' + err)
                 else:
                     out = self.get_mod_rev(self.dst_dir + '/' + model)
 
                 real_model_name_revision = out.rstrip()
                 if real_model_name_revision != '':
-                    real_model_revision = real_model_name_revision.split('@')[1][0:10]
+                    if '@' in real_model_name_revision:
+                        real_model_revision = '@' + real_model_name_revision.split('@')[1][0:10]
+                    else:
+                        real_model_revision = ''
                     real_model_name = real_model_name_revision.split('@')[0]
-                    real_model_name_revision = real_model_name + '@' + real_model_revision
+                    real_model_name_revision = real_model_name + real_model_revision
                     if force_revision_regexp:
                         missing_revision_symbol = ''
                     else:
@@ -189,7 +233,7 @@ class YangModuleExtractor:
                                 switch_items = True
 
                             # check for model revision if correct
-                            if real_model_revision != existing_model_revision:
+                            if real_model_revision.strip('@') != existing_model_revision:
                                 self.error(existing_model_name + ' model revision ' + existing_model_revision
                                            + ' is wrong or has incorrect format')
                                 switch_items = True
@@ -268,8 +312,8 @@ class YangModuleExtractor:
         ncnt = 0
         output_model = []
         for ln in input_model:
-            if ln[0].strip(' \n\r') is '':
-                if ncnt is 0:
+            if ln[0].strip(' \n\r') == '':
+                if ncnt == 0:
                     output_model.append(ln)
                 elif self.debug_level > 1:
                         self.debug_print_strip_msg(ln[1] - 1, ln[0])
@@ -375,7 +419,37 @@ class YangModuleExtractor:
                 return
             self.debug_print_strip_msg(model[-1][1] - 1, model[-1][0])
             model.pop()
-
+            
+    def check_edge_cases(self, example_match, in_code):
+        """
+        Checks for edge cases and set level to appropriate value.
+        
+        :param example_match: if example is matched in module name 
+        :param in_code: if module in CODE BEGINS section 
+        :return: level value
+        """
+        # skip all "not example" modules in strict-examples mode
+        if self.strict_examples and not example_match:
+            if self.parse_only_modules:
+                self.warning("Unable to parse not example module in strict-example mode")
+            return 0
+        # skip all example modules in <CODE BEGINS> section in strict-example mode 
+        elif self.strict_examples and example_match and in_code:
+            if self.parse_only_modules:
+                self.warning("Unable to parse example module in <CODE BEGINS> section in strict-example mode")
+            return 0
+        # check if we are not parsing example module in strict mode
+        elif self.strict and not self.strict_examples and example_match and not in_code:
+            if self.parse_only_modules:
+                self.warning("Unable to parse example module in strict mode")
+            return 0
+        # skip all modules outside <CODE BEGINS> section in strict mode
+        elif self.strict and not self.strict_examples and not in_code:
+            return 0
+        # enable to parse this module
+        else:
+            return 1
+            
     def extract_yang_model(self, content):
         """
         Extracts one or more YANG models from an RFC or draft text string in
@@ -451,11 +525,31 @@ class YangModuleExtractor:
                 #         level = 1
                 # else:
                 #     level = 1
-
-                # always set the level to 1; we decide whether or not
-                # to output at the end
-                if quotes == 0:
-                    level = 1
+                
+                # check for parse only modules list
+                if self.parse_only_modules:
+                    # finding current module name in the list
+                    if match.groups()[2] in self.parse_only_modules or output_file in self.parse_only_modules:
+                        # check if we are not parsing example module in strict mode
+                        level = self.check_edge_cases(example_match, in_code)
+                    # set level to 0 to skip modules not in the list
+                    else: 
+                        level = 0
+                # check for skip modules list
+                elif self.skip_modules:
+                    # finding current module name in the list
+                    if match.groups()[2] in self.skip_modules or output_file in self.skip_modules:
+                        # set level to 0 to skip this module
+                        print("\nSkipping '%s'" % match.groups()[2])
+                        level = 0
+                    else:
+                        level = self.check_edge_cases(example_match, in_code)
+                else:
+                    level = self.check_edge_cases(example_match, in_code)
+                
+                if level == 1:
+                    print("\nExtracting '%s'" % match.groups()[2])
+           
                 if (self.strict_name or not output_file) and level == 1 and quotes == 0:
                     if output_file:
                         revision = output_file.split('@')[-1].split('.')[0]
@@ -463,7 +557,6 @@ class YangModuleExtractor:
                                                                                   revision))
                         output_file = '{}@{}.yang'.format(match.groups()[2].strip('"\''), revision)
                     else:
-                        print("\nExtracting '%s'" % match.groups()[2])
                         output_file = '%s.yang' % match.groups()[2].strip('"\'')
                     if self.debug_level > 0:
                         print('   Getting YANG file name from module name: %s' % output_file)
@@ -499,14 +592,7 @@ class YangModuleExtractor:
                         else:
                             level += (counter['{'] - counter['}'])
                     if level == 1:
-                        if self.strict:
-                            if self.strict_examples:
-                                if example_match and not in_model:
-                                    self.write_model_to_file(model, output_file)
-                            elif in_model:
-                                self.write_model_to_file(model, output_file)
-                        else:
-                            self.write_model_to_file(model, output_file)
+                        self.write_model_to_file(model, output_file)
                         self.max_line_len = 0
                         model = []
                         output_file = None
@@ -542,7 +628,6 @@ class YangModuleExtractor:
                 mg = match.groups()
                 # Get the YANG module's file name
                 if mg[2]:
-                    print("\nExtracting '%s'" % match.groups()[2])
                     output_file = mg[2].strip()
                 else:
                     if mg[0] and mg[1] is None:
@@ -558,7 +643,8 @@ class YangModuleExtractor:
 
 
 def xym(source_id, srcdir, dstdir, strict=False, strict_name=False, strict_examples=False, debug_level=0,
-        add_line_refs=False, force_revision_pyang=False, force_revision_regexp=False):
+        add_line_refs=False, force_revision_pyang=False, force_revision_regexp=False, skip_modules=None,
+        parse_only_modules=None):
     """
     Extracts YANG model from an IETF RFC or draft text file.
     This is the main (external) API entry for the module.
@@ -576,6 +662,8 @@ def xym(source_id, srcdir, dstdir, strict=False, strict_name=False, strict_examp
     :param debug_level: Determines how much debug output is printed to the console
     :param force_revision_regexp: Whether it should create a <filename>@<revision>.yang even on error using regexp
     :param force_revision_pyang: Whether it should create a <filename>@<revision>.yang even on error using pyang
+    :param skip_modules: it will skip modules in this list
+    :param parse_only_modules: it will parse only modules in this list
     :return: None
     """
 
@@ -591,7 +679,8 @@ def xym(source_id, srcdir, dstdir, strict=False, strict_name=False, strict_examp
                      r'(?:/?|[/?]\S+)$', re.IGNORECASE)
     rqst_hdrs = {'Accept': 'text/plain', 'Accept-Charset': 'utf-8'}
 
-    ye = YangModuleExtractor(source_id, dstdir, strict, strict_examples, strict_name, add_line_refs, debug_level)
+    ye = YangModuleExtractor(source_id, dstdir, strict, strict_examples, strict_name, add_line_refs, debug_level, 
+                             skip_modules, parse_only_modules)
     is_url = url.match(source_id)
     if is_url:
         r = requests.get(source_id, headers=rqst_hdrs)
@@ -661,6 +750,16 @@ if __name__ == "__main__":
                         help="Optional: if True it will check if file contains correct revision in file name."
                              "If it doesnt it will automatically add the correct revision to the filename using regular"
                              " expression")
+    group = parser.add_mutually_exclusive_group()
+    group.add_argument(
+        "--parse-only-modules", nargs='+',
+        help="Optional: it will parse only modules added in the list in arguments."
+    )
+    group.add_argument(
+        "--skip-modules", nargs='+',
+        help="Optional: it will skip modules added in the list in arguments."
+    )
+    
     args = parser.parse_args()
 
     extracted_models = xym(args.source,
@@ -671,7 +770,10 @@ if __name__ == "__main__":
                            args.debug,
                            args.add_line_refs,
                            args.force_revision_pyang,
-                           args.force_revision_regexp)
+                           args.force_revision_regexp,
+                           skip_modules=args.skip_modules,
+                           parse_only_modules=args.parse_only_modules
+    )
     if len(extracted_models) > 0:
         if args.strict:
             print("\nCreated the following models that conform to the strict guidelines::")
